/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __PARAMETERGENOTYPEDECODER_H
#define __PARAMTERGENOTYPEDECODER_H

#include "GenotypeDecoder.h"

/*!	\file ParameterGenotypeDecoder.h
	\brief Interface of the genotype decoder that extract parameters from a genome.
*/

namespace Teem
{

	/**
	Genotype decoder to decode a set of parameter (no controller here). Parameter are
	located in a std::map. The decode needs no step. Here is an example of settings to
	use this decoder:
\code
// decoder settings
{
	name = "Parameter"
	parameters
	{
		{
			name = "Param 1";
			range = 0 1;
		}
		{
			name = "Param 2";
			range = -1 1;
		}
	}
}
\endcode
	\ingroup genotypeDecoders
	*/
	class ParameterGenotypeDecoder : public GenotypeDecoder
	{
	public:
		//! Constructor.
		ParameterGenotypeDecoder(const std::string &root);
		//! Virtual destructor, do nothing
		virtual ~ParameterGenotypeDecoder() { }

		//! Create a genome with the appropriate size.
		virtual Genome* createGenome();

		//! Decode the genome to the parameters. Returns always NULL.
		virtual Controller* decode(Genome *genome);

		//! Return the decoded value of a parameter.
		double getParameter(const std::string &key)
		{
			assert(parameters.find(key) != parameters.end());
			return parameters[key].value;
		}

	protected:
		//! Internal structure to describe a parameter.
		struct ParamDesc
		{
			An::Point range; //!< Range of parameter [range.x .. range.y]
			bool log; //!< Is parameter stored in logarithmic scale. If true, real_value = 10^(decoded_value)
			double value; //!< Actual value of parameter
		};

		//! Container for the parameters.
		std::map<std::string, ParamDesc> parameters;
	};
}

#endif
