#ifndef __NEURON_H
#define __NEURON_H

#include <set>
#include <map>
#include <string>

#include <an/Random.h>

#include "../../utils/Graph.h"
#include "../../utils/SpikeDelayLine.h"

/*!	\file Neuron.h
	\brief Interfaces of neurons and synapses
*/

/*!	\defgroup neurons Neurons and synapses */

namespace Teem
{
	class Synapse;
	
	/**
	Neuron and Synapse are interfaces to a generic NN implementation. See
	subclass for various implementation, like sigmoidal NN or SRM spiking
	NN. The synapse can convey a current (you can call it firing rate or
	anything, it's a continuous real value) and a spike train (that can be
	ignored by implementation.
	
	The time-step is implemented that way.
	
		1) All the synapses are updated using the "update" function. The
		implementation should be based on the state of presynaptic neuron.
		The default is simply to foward value provided by presynaptic
		neuron.
		
		2) All the neuron are updated using the "update" function. The 
		implementation should be based on the values provided by incoming
		synapses (synapse may alter value given by pre-synaptic neurons).
		The content of lastSpike and lastCurrent variable should not be
		altered, since it reflects the state of the last time-step (which
		should be kept until ALL neurons are updated).
		
		3) The state of all the neuron is committed (using "commit") to the
		output value. Output should be written to the lastSpike and
		lastCurrent variables.
	
	Decomposition into step 2 and 3 is necessary so that all neuron updates
	based on the state of the _previous_ time-step. It is up to the sub-
	class to implement the caching scheme, since it can be optimized depending
	on the implementation.
	
	NOTE: the synapse should normally not apply its weight on the value it
	returns. It is up to the post-synaptic neuron to take into account the weight
	of the synapse.
	\ingroup neurons
	*/
	class Neuron
	{
	public:
		//! Set of synapses
		typedef std::set<Synapse*> SynapseSet;
		
		SynapseSet incoming; //!< incoming synapses
		SynapseSet outgoing; //!< outgoing synapses
		TagSet tags; //!< tags for this neuron
	
	public:
		//! Constructor
		Neuron();
		//! Destructor
		virtual ~Neuron();
		
		//! Called when the network is totally constructed but before it is run.
		virtual void finalize() { }
	
		//! Return true is neuron is emitting a spike (that is a spike has
		//! been generated last time-step.
		virtual bool getSpike() const {return lastSpike;}
		//! Return the current emitted by this neuron (that is the one computed
		//! last time-step.
		virtual double getCurrent() const {return lastCurrent;}
	
		//! Set parameters.
		virtual void setParams(const std::map<std::string, double> params);
		//! Set parameters. Should be overriden by subclasses to handle more parameters.
		virtual void setParams(const std::string &param, double value);
		
		
		//! Update status of the neuron according to incoming synapse..
		virtual void update(double dt = 1.0) = 0;
		//! Commit state to lastSpike and lastCurrent.
		virtual void commit() = 0;
	
	protected:
		//! True if neuron emitted a spike last time-step.
		bool lastSpike;
		//! Current transmitted by neuron last time-step.
		double lastCurrent;
	
	};
	
	//! Interface for input neuron
	//! \ingroup neurons
	class InputNeuron : public Neuron
	{
	public:
		virtual ~InputNeuron() { }

		virtual void update(double dt = 1.0) { }
		virtual void commit() { }
		
		virtual void setParams(const std::string &param, double value);
		
		//! Set the input value. 
		virtual void setInputValue(double value) = 0;
	};
	
	//! Simple input neuron that will copy its value to its output current.
	//! \ingroup neurons
	class CurrentInputNeuron : public InputNeuron
	{
	public:
		virtual ~CurrentInputNeuron() { }
		
		virtual void setInputValue(double value) { lastCurrent = value; }
	};
	
	//! Simple input neuron that will spike if its value is non-null.
	//! \ingroup neurons
	class SimpleSpikeInputNeuron : public InputNeuron
	{
	public:
		virtual ~SimpleSpikeInputNeuron() { }
		
		virtual void setInputValue(double value) { lastSpike = value != 0.0; }
		
		//! This accessor is provided for the case where high optimization
		//! is necessary, but require to dynamic_cast the neuron to be used.
		void setSpike(bool spike) { lastSpike = spike; }
	};
	
	//! Input class that will generate a spike train with a probability that is
	//! equal to the input value.
	//! \ingroup neurons
	class FiringRateInputNeuron : public InputNeuron
	{
	public:
		virtual ~FiringRateInputNeuron() { }
		
		virtual void setInputValue(double value) { probability = value; }
		virtual void commit() { lastSpike = An::boolRand(probability); }
	
	protected:
		double probability; //!< probability to generate a spike at each step
	};
	
	/**
	Neuron that convert firing rate into a current. Can be used as a convertor
	in a mixed NN or as an ouput in a spiking NN. The firing rate is computed
	as follow the sum of all spike received during the time window from all the
	presynaptic neurons, normalized by the number of time-step in the time window
	and the number of pre-synaptic neurons.
	
	This class handle the following parameter:
		- \c timeWindow time during which rate, in time-steps
	\ingroup neurons
	*/
	class FiringRateOutputNeuron : public Neuron
	{
	public:
		//! Constructor.
		FiringRateOutputNeuron() : totalSpike(0), queueSize(20) { }
		//! Destructor.
		virtual ~FiringRateOutputNeuron() { }
		
		virtual void setParams(const std::string &param, double value);
		virtual void update(double dt = 1.0);
		virtual void commit();
		
	protected:
		std::queue<unsigned> spikes; //!< Buffer to save spikes during the timewindow.
		unsigned totalSpike; //!< total number of spikes:\n \f$\sum_{i=0}^{spikes.size()-1} spikes[i]\f$
		unsigned queueSize; //!< size of queue in second
		
		// FIXME: this is hacky, since it is dependend on step-size. It is
		// working for the dt = 1.0 case but doesnt scale well to different
		// time-step size values.
	};
	
	//! Connection between neurons.
	//! @see Neuron
	//! \ingroup neurons
	class Synapse
	{
	public:
		double weight; //!< weight of this synapse
		TagSet tags; //!< tags for this synapse
	
	public:
		//! Constructor
		Synapse();
		//! Destructor
		virtual ~Synapse();
		
		//! Initialize the synapse. Must be called once and only once, just after construction.
		virtual void init(Neuron* pre, Neuron* post, const double weight = 0);
		//! Called when the network is totally constructed but before it is run.
		virtual void finalize() { }
	
		//! Set parameters.
		virtual void setParams(const std::map<std::string, double> params);
		//! Set parameters. Should be overriden by subclasses to handle more parameters.
		virtual void setParams(const std::string &param, double value);
		
		//! Return true if there is a spike on this synapse
		virtual bool getSpike() const {return preNeuron ? preNeuron->getSpike() : false;}
		//! Return the value of the potential of this synapse
		virtual double getCurrent() const {return preNeuron ? preNeuron->getCurrent() : 0.0;}
		
		//! Update status of the synapse according to state of presynaptic neurons.
		virtual void update(double dt = 1.0) {}
		
		//! Get the pre-synaptic neuron.
		const Neuron *getPreNeuron() const { return preNeuron; }
		
		//! Get the post-synaptic neuron.
		const Neuron *getPostNeuron() const { return postNeuron; }
	
	protected:
		Neuron* preNeuron; //!< pre-synaptic neuron
		Neuron* postNeuron; //!< post-synaptic neuron
	};
}


#endif
