#include "SRMNeuron.h"
#include <an/Random.h>

/*!	\file SRMNeuron.cpp
	\brief Implementations of SRM (Spike Response Model) neurons and synapses
*/

namespace Teem
{
	using namespace An;
	
	double SRMPlasticSynapse::learningRate = 1.0;
	double SRMPlasticSynapse::stdpControl = 1.0;
	
	
	SRMNeuron::SRMNeuron()
	{
		// default value for settings
		threshold = 1.0;
		thresNoise = 0.01;
		refracQuantum = 1.0;
		
		// state
		membrane = 0.0;
		refractory.setTau(4.0);
		isSpiking = false;
		
		// we never output current.
		lastCurrent = 0.0;
	}
	
	void SRMNeuron::setParams(const std::string &param, double value)
	{
		if (param == "threshold")
			threshold = value;
		else if(param == "refracQuantum")
			refracQuantum = value;
		else if(param == "tau")
			refractory.setTau(value);
		else if(param == "thresholdNoise")
			thresNoise = value;
		else
			Neuron::setParams(param, value);
	}
	
	void SRMNeuron::update(double dt)
	{
		membrane = 0.0;
		
		// integrate on synapse
		for (SynapseSet::iterator p = incoming.begin(); p != incoming.end(); ++p)
			membrane += (*p)->getCurrent();
		
		// add refractoryness
		membrane += (double) refractory;
		
		// step
		refractory.step(dt);
		
		// is spiking
		if (membrane >= threshold + gaussianRand(0.0, threshold * thresNoise))
		{
			isSpiking = true;
			refractory -= refracQuantum;
		}
		else
			isSpiking = false;
	}
	
	void SRMNeuron::commit()
	{
		lastSpike = isSpiking;
	}
	
	void DelaySynapse::setParams(const std::string &param, double value)
	{
		if(param == "delay")
			setDelay(value);
		else
			Synapse::setParams(param, value);
	}
	
	void DelaySynapse::update(double dt)
	{
		// step the delay line.
		incomingSpikes = delayLine.step(dt, preNeuron ? preNeuron->getSpike() : false);
	}
	
	SRMSynapse::SRMSynapse()
	{	
		setTau(10.0);
		response = 0.0;
	}
	
	void SRMSynapse::setParams(const std::string &param, double value)
	{
		if(param == "tau")
			setTau(value);
		else
			DelaySynapse::setParams(param, value);
	}
	
	void SRMSynapse::update(double dt)
	{
		// call inherited
		DelaySynapse::update(dt);
		
		// handle incoming spike.
		spikeResponse1 += ((double) incomingSpikes) * weight;
		spikeResponse2 += ((double) incomingSpikes) * weight;
		
		// compute response.
		response =  4.0 * ((double) spikeResponse1 - (double) spikeResponse2);
		
		// step
		spikeResponse1.step(dt);
		spikeResponse2.step(dt);
	}
	
	SRMPlasticSynapse::SRMPlasticSynapse()
	{
		potFactor = 0.1;
		deprFactor = 0.12;
		potState.setTau(4.0);
		deprState.setTau(4.0);
		preDelayLine.setDelay(1.0);
		postDelayLine.setDelay(1.0);
	}
	
	void SRMPlasticSynapse::setParams(const std::string &param, double value)
	{
		if(param == "potFactor")
			potFactor = value;
		else if(param == "deprFactor")
			deprFactor = value;
		else if(param == "potTau")
			potState.setTau(value);
		else if(param == "deprTau")
			deprState.setTau(value);
		else if(param == "learningDelay")
		{
			preDelayLine.setDelay(value);
			postDelayLine.setDelay(value);
		}
		else
			SRMSynapse::setParams(param, value);
	}
	
	void SRMPlasticSynapse::update(double dt)
	{	
		SRMSynapse::update(dt);
		preformLearning(dt);
	}
	
	void SRMPlasticSynapse::preformLearning(double dt)
	{	
		// step delay lines
		unsigned preSpiked = preDelayLine.step(dt, incomingSpikes > 0);
		unsigned postSpiked = postDelayLine.step(dt, postNeuron ? postNeuron->getSpike() : false);
		
		// update state variables
		if(preSpiked)
			potState += potFactor;
		if(postSpiked)
			deprState -= deprFactor;
		
		// compute dw
		double dw = 0.0;
		if(postNeuron && postNeuron->getSpike())
			dw += (double) potState;
		if(incomingSpikes)
			dw += (double) deprState;
		
		// step state variables
		potState.step(dt);
		deprState.step(dt);
		
		// damping - only simple clipping implemented yet
		weight += stdpControl * learningRate * dw;
		if(weight > 1.0)
			weight = 1.0;
		else if(weight < 0.0)
			weight = 0.0;
	}
}
