#ifdef HAVE_VIEWER

#include "EnkiViewer.h"
#include <enki/robots/s-bot/SbotObject.h>
#include <enki/robots/s-bot/Sbot.h>
#include <enki/robots/khepera/Khepera.h>
#include <enki/robots/alice/Alice.h>
#include <enki/PhysicalEngine.h>
#include <an/Geometry.h>
#include "../world/Enki.h"

#include <assert.h>

using namespace Enki;
using namespace An;
//using namespace Teem;

Viewer::Viewer(bool recVideo) :
	realtime("teem.realtime", 0),
	fontFileName("teem.viewer.fontFileName", "sans.ttf")
{
	displayAll = false;
	eventStep = 0;
	
	#ifdef HAVE_AVCODEC
	if (recVideo)
		vrec = new VideoRecorder();
	else
		vrec = NULL;
	#endif
	font = NULL;
}

Viewer::~Viewer()
{
	#ifdef HAVE_AVCODEC
	if (vrec)
		delete vrec;
	#endif
	if (font)
		delete font;
	delete gc;
}

void Viewer::init(int width, int height)
{
	this->width = width;
	this->height = height;
	gc = GAG::GraphicContext::createGraphicContext();
	gc->setRes(width, height, 32, GAG::DrawableSurface::DOUBLEBUF);
	gc->setCaption("EnkiViewer", "EnkiViewer");
	font = gc->loadFont(fontFileName, 10);
}

void Viewer::step(double dt)
{
	if (realtime)
	{
		startDelayFrame();
		display(true);
		endDelayFrame(dt);
	}
	else
	{
		display(false);
	}
	handleEvent(true);
}

void Viewer::startRecordingVideo(const std::string &fileName)
{
	#ifdef HAVE_AVCODEC
	if (vrec)
		vrec->openFile(fileName.c_str(), gc->getW(), gc->getH());
	#endif
}

void Viewer::stopRecordingVideo(void)
{
	#ifdef HAVE_AVCODEC
	if (vrec)
		vrec->closeFile();
	#endif
}

void Viewer::handleEvent(bool eachStep)
{
	if (displayAll || eachStep || (((eventStep++) & 0xff)==0))
	{
		// handle event
		SDL_Event event;
		while (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_QUIT:
				{
					#ifdef HAVE_AVCODEC
					if (vrec)
						vrec->closeFile();
					#endif
					exit(0);
				}
				break;
				
				case SDL_MOUSEBUTTONUP:
				{
					displayAll = !displayAll;
				}
				break;
				
				default:
				break;
			}
		}
	}
}

void Viewer::display(bool force)
{
	if ((!displayAll) && (!force))
		return;
		
	displayViewer(width, height);

	#ifdef HAVE_AVCODEC
	if (vrec)
		vrec->encodeFrame((Uint8 *)gc->getPixelPointer());
	#endif

	gc->nextFrame();
}

void Viewer::setCaption(const std::string &caption)
{
	gc->setCaption(caption.c_str(), "Steph's cool viewer");
}

void Viewer::startDelayFrame(void)
{
	delayStartTick = SDL_GetTicks();
}

void Viewer::endDelayFrame(double dt)
{
	assert(dt>=0);

	unsigned elapsed = SDL_GetTicks()-delayStartTick;
	unsigned duration = static_cast<unsigned>(dt*1000);

	if (elapsed<duration)
		SDL_Delay(duration-elapsed);
}


EnkiViewer::EnkiViewer(const Enki::World *w, bool recVideo, int sideBarWidth) :
	Viewer(recVideo),
	scale("teem.viewer.scale", 0.0)
{
	this->world = w;
	
	double minSize = std::min(world->w, world->h);
	if (scale == 0.0)
	{
		if (minSize < 200)
			scale = 200.0 / minSize;
		else
			scale = 1.0;
	}
	this->sideBarWidth = sideBarWidth;
	init(static_cast<int>((world->w + 20)*scale + sideBarWidth), static_cast<int>((world->h + 20)*scale));
}

void EnkiViewer::displayViewer(int w, int h)
{
	// fill screen
	gc->drawFilledRect(0, 0, gc->getW(), gc->getH(), 127, 127, 127);

	// draw world
	displayWorld(0, 0, (int)((world->w + 20)*scale), (int)((world->h + 20)*scale));

	// draw side bar
	displaySideBar((int)((world->w + 20)*scale), 0, sideBarWidth, (int)((world->h + 20)*scale));
}


void EnkiViewer::displayWorld(int, int, int, int)
{
	// draw walls
	gc->drawRect( static_cast<int>(10*scale),  static_cast<int>(10*scale), static_cast<int>(world->w*scale), static_cast<int>(world->h*scale), 0, 0, 0);
	
	// draw sound
	if (SbotGlobalSound::getWorldFrequenciesState() & 0x1)
		gc->drawFilledRect(10, 2, 20, 4, 255, 0, 0);
	if (SbotGlobalSound::getWorldFrequenciesState() & 0x2)
		gc->drawFilledRect(40, 2, 20, 4, 255, 0, 0);
	
	// draw world
	int bottom = static_cast<int>((world->w+20)*scale);
	//int yStartPos = 10;
	for (unsigned i=0; i<world->objects.size(); i++)
	{
		// draw objects
		Color c = world->objects[i]->color;
		if (world->objects[i]->boundingSurface)
		{
			const Polygone &bs = world->objects[i]->getTrueBoundingSurface();
			for (unsigned j=0; j<bs.size(); j++)
			{
				double x1 = bs[j].x+10;
				double y1 = bs[j].y+10;
				double x2 = bs[(j+1)%bs.size()].x+10;
				double y2 = bs[(j+1)%bs.size()].y+10;
				gc->drawLine(static_cast<int>(x1*scale), bottom - static_cast<int>(y1*scale), static_cast<int>(x2*scale), bottom - static_cast<int>(y2*scale), static_cast<Uint8>(255.0*c.components[0]), static_cast<Uint8>(255.0*c.components[1]), static_cast<Uint8>(255.0*c.components[2]));
			}
			double x = world->objects[i]->pos.x+10;
			double y = world->objects[i]->pos.y+10;
			double r = 3;
			double a = world->objects[i]->angle;
			
			gc->drawLine(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>((x+cos(a)*r)*scale), bottom - static_cast<int>((y+sin(a)*r)*scale), 0, 0, 0);
			
			// if Alice draw color circle inside
			Alice *ea = dynamic_cast<Alice *>(world->objects[i]);
			if (ea && font)
				gc->drawString(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), font, i);
		}
		else
		{
			double x = world->objects[i]->pos.x+10;
			double y = world->objects[i]->pos.y+10;
			double r = world->objects[i]->r;
			double a = world->objects[i]->angle;
			gc->drawCircle(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>(r*scale), static_cast<Uint8>(255.0*c.components[0]), static_cast<Uint8>(255.0*c.components[1]), static_cast<Uint8>(255.0*c.components[2]));
			gc->drawLine(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>((x+cos(a)*r)*scale), bottom - static_cast<int>((y+sin(a)*r)*scale), 0, 0, 0);
			SbotActiveObject *ao = dynamic_cast<SbotActiveObject *>(world->objects[i]);
			if (ao)
			{
				if (ao->feeding.actualTime < ao->feeding.activeDuration)
				{
					if (ao->feeding.dEnergyActive > 0)
					{
						gc->drawCircle(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>(r*scale*0.5), 0, 255, 0);
					}
					else if (ao->feeding.dEnergyActive < 0)
					{
						gc->drawCircle(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>(r*scale*0.5), 255, 0, 0);
					}
				}
				else
				{
					if (ao->feeding.dEnergyInactive > 0)
					{
						gc->drawCircle(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>(r*scale*0.5), 0, 255, 0);
					}
					else if (ao->feeding.dEnergyInactive < 0)
					{
						gc->drawCircle(static_cast<int>(x*scale), bottom - static_cast<int>(y*scale), static_cast<int>(r*scale*0.5), 255, 0, 0);
					}
				}
			}
		}
		
		// object color for sbot objects
		SbotActiveObject *ao = dynamic_cast<SbotActiveObject *>(world->objects[i]);
		if (font && ao)
		{
			std::ostringstream oss;
			oss.precision(3);
			oss << c.components[0] << " " << c.components[1] << " " << c.components[2];
			double x = world->objects[i]->pos.x+10;
			double y = world->objects[i]->pos.y+10;
			double r = world->objects[i]->r;
			gc->drawString(static_cast<int>((x + r) * scale), bottom - static_cast<int>((y + r)* scale) - 10, font, oss.str().c_str());
		}
		
		// Markus Sensor rays
		// with Antoine's line:
		Alice *ea = dynamic_cast<Alice *>(world->objects[i]);
		if (ea && i==0)
		{
			
 			drawIRSensors (bottom, ea, 0);
 			drawIRSensors (bottom, ea, 1);
 			drawIRSensors (bottom, ea, 2);
 			drawIRSensors (bottom, ea, 3);
 			drawIRSensors (bottom, ea, 4);
// 			drawCameraRays (bottom, ea);
		}
	}
}

void EnkiViewer::displaySideBar(int x, int y, int w, int h)
{
}

// Markus' IRSensor visualizer
void EnkiViewer::drawIRSensors (int bottom, const Enki::Alice *owner, int which)
{
	IRSensor sensor = owner->left;
	switch (which)
	{
		case 0: sensor = owner->left; break;
		case 1: sensor = owner->front; break;
		case 2: sensor = owner->right; break;
		case 3: sensor = owner->back; break;
		case 4: sensor = owner->frontHighExt; break;
		sensor = owner->front;
	}

	double xx = sensor.getAbsolutePosition().x + 10;
	double yy = sensor.getAbsolutePosition().y + 10;
	unsigned rs = sensor.getRayCount();
	
	// true angle
	double aa = sensor.getAbsoluteOrientation() - sensor.getAperture();
	// increment to sensor angle between rays
	double inc=0;
	if (rs > 1)
		inc = 2*sensor.getAperture() / (rs - 1.0);

	// draw all rs sensor rays
	for (unsigned i = 0; i < rs; i++)
	{
		gc->drawLine(static_cast<int>(xx*scale), bottom - static_cast<int>(yy*scale), static_cast<int>((xx+cos(aa)*sensor.getRange())*scale), bottom - static_cast<int>((yy+sin(aa)*sensor.getRange())*scale), 0, 0, 255);
		aa += inc;
	}
	// also draw smart radius
	gc->drawCircle(static_cast<int>((sensor.getAbsSmartPos().x+10)*scale), bottom - static_cast<int>((sensor.getAbsSmartPos().y+10)*scale), static_cast<int>(sensor.getSmartRadius()*scale), 0, 0, 255);
}

// Markus' Camera visualizer
void EnkiViewer::drawCameraRays (int bottom, const Enki::Alice *owner)
{
	CircularCam camera = owner->circcam;

	double xx = camera.getAbsolutePosition().x+10;
	double yy = camera.getAbsolutePosition().y+10;
	unsigned rs = camera.zbuffer.size();
	for (unsigned i = 0; i < rs; i++)
	{
		// true angle
		double aa = camera.getAbsoluteOrientation() - camera.fieldOfView + i*camera.fieldOfView*2/(rs-1);
		gc->drawLine(static_cast<int>(xx*scale), bottom - static_cast<int>(yy*scale), static_cast<int>((xx+cos(aa)*100)*scale), bottom - static_cast<int>((yy+sin(aa)*100)*scale), 255, 0, 0);
	}
}


// this is most hacky since it doesnt use the parameter to ensure position.
void VisionEnkiViewer::displaySideBar(int x, int y, int w, int h)
{
	int bottom = static_cast<int>((world->w+20)*scale);
	int yStartPos = y;
	int xStartPos = x;
	for (unsigned i=0; i < world->objects.size(); i++)
	{
		// If it is a sbot
		FeedableSbot *sbot = dynamic_cast<FeedableSbot *>(world->objects[i]);
		if (sbot)
		{
			// draw camera
			yStartPos += 40;
			
			unsigned sz = sbot->camera.zbuffer.size();
			for (unsigned j=0; j<sz; j++)
			{
				unsigned idx = sz-j-1;
				double dist = sqrt(sbot->camera.zbuffer[idx]);
				int len = (int)(400 / (dist));
				gc->drawVertLine(xStartPos+j, yStartPos-40, 40, 160, 160, 160);
				gc->drawVertLine(xStartPos+j, yStartPos+8, 40, 60, 60, 60);
				gc->drawVertLine(xStartPos+j, yStartPos-len, 8+2*len, (Uint8)(sbot->camera.image[idx].components[0]*255), (Uint8)(sbot->camera.image[idx].components[1]*255), (Uint8)(sbot->camera.image[idx].components[2]*255));
			}
			
			yStartPos += 48;
			
			yStartPos += 10;
			
			if (yStartPos + 88 > y + h)
			{
				yStartPos = y;
				xStartPos += sz + 10;
			}
		}
		
		// If it is a khepera
		Khepera *khep = dynamic_cast<Khepera *>(world->objects[i]);
		if (khep)
		{
			double xx = khep->pos.x+10;
			double yy = khep->pos.y+10;
			double aa = khep->angle;
			double fov = khep->camera.fieldOfView;
			gc->drawLine(static_cast<int>(xx*scale), bottom - static_cast<int>(yy*scale), static_cast<int>((xx+cos(aa+fov)*50)*scale), bottom - static_cast<int>((yy+sin(aa+fov)*50)*scale), 255, 0, 0);
			gc->drawLine(static_cast<int>(xx*scale), bottom - static_cast<int>(yy*scale), static_cast<int>((xx+cos(aa-fov)*50)*scale), bottom - static_cast<int>((yy+sin(aa-fov)*50)*scale), 255, 0, 0);
			
			drawIRSensor(1, 2.5, M_PI/2, bottom, *khep);
			drawIRSensor(2, 2, M_PI/4, bottom, *khep);
			drawIRSensor(2.8, 0.6, 0, bottom, *khep);
			drawIRSensor(2.8, -0.6, 0, bottom, *khep);
			drawIRSensor(2, -2, -M_PI/4, bottom, *khep);
			drawIRSensor(1, -2.5, -M_PI/2, bottom, *khep);
			drawIRSensor(-2.7, -1, -M_PI, bottom, *khep);
			drawIRSensor(-2.7, 1, -M_PI, bottom, *khep);
			
			gc->setClipRect(x, y+20, 128, 160);
			yStartPos += 80;
			// draw camera
			unsigned sz = khep->camera.zbuffer.size();
			for (unsigned j=0; j<sz; j++)
			{
				unsigned idx = sz-j-1;
				double dist = sqrt(khep->camera.zbuffer[idx]);
				int len = (int)(400 / (dist));
				gc->drawVertLine(x+j*2, yStartPos-80, 80, 160, 160, 160);
				gc->drawVertLine(x+j*2, yStartPos+8, 80, 60, 60, 60);
				gc->drawVertLine(x+j*2+1, yStartPos-80, 80, 160, 160, 160);
				gc->drawVertLine(x+j*2+1, yStartPos+8, 80, 60, 60, 60);
				gc->drawVertLine(x+j*2, yStartPos-len, 8+2*len, (Uint8)(khep->camera.image[idx].components[0]*255), (Uint8)(khep->camera.image[idx].components[1]*255), (Uint8)(khep->camera.image[idx].components[2]*255));
				gc->drawVertLine(x+j*2+1, yStartPos-len, 8+2*len, (Uint8)(khep->camera.image[idx].components[0]*255), (Uint8)(khep->camera.image[idx].components[1]*255), (Uint8)(khep->camera.image[idx].components[2]*255));
			}
			gc->setClipRect();
		}
	}
}

void VisionEnkiViewer::drawIRSensor(double dx, double dy, double angle, int bottom, const Enki::PhysicalObject &owner)
{
	Matrix22 rot(owner.angle);
	Point pos = owner.pos + rot * Point(dx, dy) + Point(10, 10);
	double a = owner.angle + angle;
	gc->drawLine(static_cast<int>(pos.x*scale), static_cast<int>(bottom - pos.y*scale), static_cast<int>(scale*(pos.x+cos(a)*3)), static_cast<int>(bottom - scale*(pos.y+sin(a)*3)), 0, 128, 0);
}

#endif // HAVE_VIEWER

