#ifdef HAVE_VIEWER
#include "NeuralNetworkDrawer.h"


#include <an/Geometry.h>

namespace Teem
{
	NeuralNetworkDrawer::NeuralNetworkDrawer(GAG::GraphicContext *gc, int x, int y, int w, int h, Controller* nn, double colorScale)
	{
		this->gc = gc;
		this->x = x;
		this->y = y;
		this->w = w;
		this->h = h;
		this->colorScale = colorScale;
		
		FeedForwardNeuralNetwork* ffnn = dynamic_cast<FeedForwardNeuralNetwork*>(nn);
		if(ffnn != NULL)
			drawFeedForwardNeuralNetwork(ffnn);
	}
	
	NeuralNetworkDrawer::NeuralNetworkDrawer(GAG::GraphicContext *gc, int x, int y, int w, int h, double colorScale)
	{
		this->gc = gc;
		this->x = x;
		this->y = y;
		this->w = w;
		this->h = h;
		this->colorScale = colorScale;

		cellSize = 12;
	}
	
	//! Returns the min of a and b, when they are not the same class and require casting
	template<class A, class B> inline A min(A a, B b) { return a < b ? a : b; }
	
	void NeuralNetworkDrawer::drawFeedForwardNeuralNetwork(FeedForwardNeuralNetwork* ffnn)
	{
		// init geometry
		unsigned numFromNeurons;
		unsigned numToNeurons = 0;
		for(unsigned i = 0; i < ffnn->layerNum(); i++)
			numToNeurons += ffnn->layerSize(i);
		numFromNeurons = numToNeurons + 1 + ffnn->inputNum();
		
		// cell size
		int hz = std::min(12, (int)(w/(numFromNeurons+1)));
		int vz = std::min(12, (int)(h/(numToNeurons+1)));
		cellSize = std::min(vz, hz);
		
		// draw header
		for(unsigned i = 1; i <= numFromNeurons; i++)
			drawCell(i, 0, true);
		for(unsigned i = 1; i <= numToNeurons; i++) 
			drawCell(0, i, true);
		
		// for each layer, draw the connection matrix
		unsigned offX = 1;
		unsigned offY = 1;
		unsigned totW = (numFromNeurons+1)*cellSize, totH = (numToNeurons+1)*cellSize;
		for(unsigned layer = 0; layer < ffnn->layerNum(); layer++)
		{
			unsigned fromCount = (layer == 0 ? ffnn->inputNum() : ffnn->layerSize(layer-1));
			unsigned toCount = ffnn->layerSize(layer);
			
			// draw weight matrix
			for(unsigned i = 0; i < fromCount; i++)
				for(unsigned j = 0; j < toCount; j++)
					drawCell(offX + i, offY + j, false, ffnn->getWeight(layer, i, j));
			
			// draw the bias weights
			for(unsigned j = 0; j < toCount; j++)
				drawCell(1 + ffnn->inputNum(), offY + j, false, ffnn->getBiasWeight(layer, j));
			
			// shift position
			offX += fromCount;
			offY += toCount;
			
			// draw separation line
			if(layer != ffnn->layerNum()-1)
				gc->drawHorzLine(x, y + offY*cellSize, totW, 0, 0, 0);
			gc->drawVertLine(x + offX*cellSize, y, totH, 0, 0, 0);
			
			// accomodate room for bias column
			if(layer == 0)
			{
				++offX;
				gc->drawVertLine(x + offX*cellSize, y, totH, 0, 0, 0);
			}
		}
		
		// check for recurrency
		RecurrentFeedForwardNeuralNetwork* rffnn = dynamic_cast<RecurrentFeedForwardNeuralNetwork*>(ffnn);
		if(rffnn != NULL)
			drawRecurrencyFromFeedForwardNeuralNetwork(rffnn);
	}
	
	void NeuralNetworkDrawer::drawRecurrencyFromFeedForwardNeuralNetwork(RecurrentFeedForwardNeuralNetwork* rffnn)
	{
		unsigned offX = 1 + rffnn->inputNum() + 1;
		unsigned offY = 1;
		for(unsigned layer = 0; layer < rffnn->layerNum(); layer++)
		{
			if(rffnn->isLayerRecursive(layer))
			{
				for(unsigned i = 0; i < rffnn->layerSize(layer); i++)
					for(unsigned j = 0; j < rffnn->layerSize(layer); j++)
						drawCell(offX+i, offY+j, false, rffnn->getRecursiveWeight(layer, i, j));
			}
			
			offX += rffnn->layerSize(layer);
			offY += rffnn->layerSize(layer);
		}
	}
		
	void NeuralNetworkDrawer::drawCell(int cx, int cy, bool round, double color)
	{
		int sz = cellSize - 3;
		int xx = cx * cellSize+2;
		int yy = cy * cellSize+2;
		
		if(round)
		{
			gc->drawCircle(x + xx + sz/2, y + yy + sz/2, sz/2, 0, 0, 0);
		}
		else
		{
			int cc = static_cast<int>(color / colorScale * 255.0);
			if(cc > 255)
				cc = 255;
			else if(cc < -255)
				cc = -255;
			int r, g, b;
			if(cc > 0)
			{
				r = 255; g = 255 - cc; b = 255 - cc;
			}
			else
			{
				r = 255 + cc; g = 255 + cc; b = 255;
			}
			
			gc->drawFilledRect(x + xx, y + yy, sz, sz, r, g, b);
			gc->drawRect(x + xx, y + yy, sz, sz, 0, 0, 0);
		}
	}
}

#endif
