#ifdef HAVE_ENKI

#include "BatteryRecharge.h"
#include "../utils/Logger.h"
#include "../controller/Joystick.h"
#include "../world/Enki.h"
#include "../main/VerboseLevel.h"
#include "../controller/Controller.h"
#include "../controller/GenotypeDecoder.h"


// HACK
#ifdef HAVE_VIEWER
#include "../enkiviewer/EnkiViewer.h"
#include "../enkiviewer/NeuralNetworkDrawer.h"
#endif

#include <an/Random.h>

#include <sstream>



namespace Teem
{
	#ifdef HAVE_VIEWER
	
	class BatteryRechargeEnkiViewer : public VisionEnkiViewer
	{
	protected:
		BatteryRecharge* exp;
	public:
		BatteryRechargeEnkiViewer(const Enki::World *w, bool recVideo, BatteryRecharge* exp)
			: VisionEnkiViewer(w, recVideo, 400)
		{
			this->exp = exp;
		}
		virtual ~BatteryRechargeEnkiViewer() { }
		
	protected:
		virtual void displayWorld(int x, int y, int w, int h)
		{
			int bottom = static_cast<int>((world->w+20)*scale);
			
			// draw recharge zone.	
			gc->setClipRect(10*scale, bottom - 10*scale - exp->rechargeZoneRadius*scale, exp->rechargeZoneRadius*scale, exp->rechargeZoneRadius*scale);
			gc->drawCircle(10*scale, bottom - 10*scale, exp->rechargeZoneRadius*scale, 0, 0, 0);
			gc->setClipRect();
			
			VisionEnkiViewer::displayWorld(x, y, w, h);
		}
		
		virtual void displaySideBar(int x, int y, int w, int h)
		{
			y += 15;
			VisionEnkiViewer::displaySideBar(x, y, w, h);
			
			// draw IR values.
			int startX = x + 120;
			int startY = y + 30;
			int dw = 8;
			int dh = 40;
			int dx[] = {0, 2*dw, 4*dw, 6*dw, 8*dw, 10*dw, 6*dw, 4*dw};
			int dy[] = {dh, dh/2, 0, 0, dh/2, dh, dh*2, dh*2}; 
			for(unsigned i = 0; i < 8; i++)
			{
				gc->drawRect(startX + dx[i], startY + dy[i], dw, dh, 0, 0, 0);
				double d = dh * exp->ir[i];
				gc->drawFilledRect(startX + dx[i], startY + dy[i] + dh - d, dw, d, 255, 255, 255);
			}
			
			// draw light sensors
			startX = x + 230;
			startY = y + 45;
			int c = exp->frontAmbiant * 255.0;
			gc->drawFilledRect(startX, startY, 30, 30, c, c, c);
			c = exp->backAmbiant * 255.0;
			gc->drawFilledRect(startX, startY + 50, 30, 30, c, c, c);
			
			// draw battery level
			startX = x + 290;
			startY = y + 40;
			dw = 10;
			dh = 100;
			int bat = dh * exp->batteryLevel / exp->batteryCapacity;
			gc->drawRect(startX, startY, dw, dh, 0, 0, 0);
			gc->drawFilledRect(startX, startY + dh - bat, dw, bat, 255, 255, 255);
			
			// draw floor sensor
			startX = x + 320;
			startY = y + 55;
			c = exp->floorSensor * 255.0;
			gc->drawFilledRect(startX, startY, 30, 30, c, c, c);
			
			// draw neural network
			NeuralNetworkDrawer(gc, x, y + 200, 1000, 1000, exp->controller, 2.0);
		}
	};
	
	#endif
	
	REGISTER_EXPERIMENT(BatteryRecharge);
	
	BatteryRecharge::BatteryRecharge() :
		inputIR("teem.experiment.inputIR", true),
		inputEncoders("teem.experiment.inputEncoders", true),
		inputBatteryLevel("teem.experiment.inputBatteryLevel", true),
		inputAmbiant("teem.experiment.inputAmbiant", true),
		inputFloor("teem.experiment.inputFloor", true),
		batteryCapacity("teem.experiment.batteryCapacity", 20.0),
		initialBatteryCapacity("teem.experiment.initialBatteryCapacity", 20.0),
		maxSpeed("teem.experiment.maxSpeed", 6.0),
		rechargeZoneRadius("teem.experiment.rechargeZoneRadius", 10.0),
		fitnessFunction("teem.experiment.fitnessFunction", "0.5 * (enc_l + enc_r)"),
		fitnessFormula(fitnessFunction)
	{
		ir.resize(8);
		joystick = false;
		
		// Setup fitness formula.
		fitnessFormula.setVariable("enc_l", leftEncoder);
		fitnessFormula.setVariable("enc_r", rightEncoder);
		fitnessFormula.setVariable("mot_l", leftMotor);
		fitnessFormula.setVariable("mot_r", rightMotor);
		fitnessFormula.setVariable("x", pos.x);
		fitnessFormula.setVariable("y", pos.y);
		fitnessFormula.setVariable("orient", orientation);
		fitnessFormula.setVariable("bat", batteryLevel);
		fitnessFormula.setVariable("amb_f", frontAmbiant);
		fitnessFormula.setVariable("amb_b", backAmbiant);
		fitnessFormula.setVariable("floor", floorSensor);
		for(unsigned i = 0; i < 8; i++)
		{
			std::ostringstream oss; oss << "ir_" << i;
			fitnessFormula.setVariable(oss.str(), ir[i]);
		}
		
		fitnessFormula.parse();
		
		// Create and setup world and robot.
		EnkiWorld* enkiWorld = new EnkiWorld(false);
		#ifdef HAVE_VIEWER
		enkiWorld->createViewer<BatteryRechargeEnkiViewer>(this);
		#endif
		w = enkiWorld;
		assert(w->getCapability<KheperaCreator>());
		robot = w->getCapability<KheperaCreator>()->createKhepera();
	}
		
	void BatteryRecharge::setupDecoders(GenotypeDecoders &decoders)
	{
		// count the number of inputs we are going to use
		unsigned numInput = 0;
		for(unsigned i = 0; i < 8; i++)
			if(inputIR[i])
				numInput++;
		if(inputEncoders[0]) numInput++;
		if(inputEncoders[1]) numInput++;
		if(inputBatteryLevel) numInput++;
		if(inputAmbiant[0]) numInput++;
		if(inputAmbiant[1]) numInput++;
		if(inputFloor) numInput++;
		
		// we assume we are using a single decoder, since we are in a simple evolution setup
		// also, we have only 2 input for the motors
		assert(decoders.size() == 1);
		decoders[0]->controllerIOSetup(numInput, 2);
	}
	
	void BatteryRecharge::createControllers(GenotypeDecoders &decoders, GenerationalPopulation &individuals, unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{
		assert(decoders.size() == 1);
		assert(individuals.size() == 1);
		
		controller = decoders[0]->decode(individuals[0].genome);
	}
	
	void BatteryRecharge::initEpoch(unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{
		batteryLevel = initialBatteryCapacity;
		
		// init robot position
		robot->setPosition(An::Point(An::uniformRand() * 60.0, An::uniformRand()*60.0));
		robot->setOrientation(An::uniformRand() * M_PI * 2.0);
		
		// setup logging
		if(mode == MODE_TEST_INDIVIDUAL)
		{
			std::ostringstream oss;
			oss << "epoch_ind" << ind << "_gen" << gen << ".log";
			logger = new OctaveLogger(oss.str().c_str());
			logger->addVariable("leftEnc", leftEncoder);
			logger->addVariable("rightEnc", rightEncoder);
			logger->addVariable("leftMot", leftMotor);
			logger->addVariable("rightMot", rightMotor);
			logger->addVariable("x", pos.x);
			logger->addVariable("y", pos.y);
			logger->addVariable("orient", orientation);
			logger->addVariables("ir", &ir[0], &ir[ir.size()]);
			logger->addVariable("frontAmb", frontAmbiant);
			logger->addVariable("backAmb", backAmbiant);
			logger->addVariable("floor", floorSensor);
			logger->addVariable("bat", batteryLevel);
		}
		else
			logger = NULL;
	}
	
	void BatteryRecharge::termEpoch(unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{
		if(logger != NULL)
		{
			delete logger;
			logger = NULL;
		}
		
		delete controller;
	}
	
	double BatteryRecharge::doSensoryMotorCycle(unsigned cycle, bool &kill)
	{
		assert(robot != NULL);
		assert(controller != NULL || joystick);
		
		double fitness = 0.0;
		kill = false;
		
		// read sensors
		for(unsigned i = 0; i < 8; i++)
			ir[i] = robot->getIRSensor(i) / 1100.0;
		robot->getEncoders(leftEncoder, rightEncoder);
		pos = robot->getPosition();
		orientation = robot->getOrientation();
		leftEncoder /= maxSpeed;
		rightEncoder /= maxSpeed;
		
		// update battery level
		if(pos.norm2() < rechargeZoneRadius * rechargeZoneRadius)
		{
			batteryLevel = batteryCapacity;
			floorSensor = 0.0;
		}
		else
		{
			batteryLevel -= simulationStep;
			floorSensor = 1.0;
			if(batteryLevel <= 0.0)
			{
				kill = true;
				batteryLevel = 0.0; // cap to 0.0 for when using the joystick
			}
		}
		
		// simulate ambiant sensors
		frontAmbiant = simulateAmbiant(2.8, 0.6, 0);
		backAmbiant = simulateAmbiant(-2.7, -1.0, -M_PI);
		
		// compute the fitness
		fitness = computeFitness();
		
		// compute motor value
		if(!joystick)
		{
			assignInputsToController();
			controller->step();
			leftMotor = controller->getOutput(0);
			rightMotor = controller->getOutput(1);
		}
		else
		{
			leftMotor = leftJoystick;
			rightMotor = rightJoystick;
		}
		
		// assign motor value and step
		robot->setSpeed(maxSpeed * leftMotor, maxSpeed * rightMotor);
		w->step(simulationStep);
		
		// log
		if(logger != NULL)
			logger->step();
		
		return fitness;
	}
	
	double BatteryRecharge::computeFitness()
	{
		return fitnessFormula.eval();
	}
	
	void BatteryRecharge::assignInputsToController()
	{
		assert(controller != NULL);
		assert(robot != NULL);
		
		// ir sensors
		unsigned index = 0;
		for(unsigned i = 0; i < 8; i++)
			if(inputIR[i])
				controller->setInput(index++, ir[i]);
		
		// encoders
		if(inputEncoders[0]) controller->setInput(index++, leftEncoder);
		if(inputEncoders[1]) controller->setInput(index++, rightEncoder);
		
		// battery level
		if(inputBatteryLevel) controller->setInput(index++, batteryLevel);
		
		// ambiant light
		if(inputAmbiant[0]) controller->setInput(index++, frontAmbiant);
		if(inputAmbiant[1]) controller->setInput(index++, backAmbiant);
		
		// floor sensor
		if(inputFloor) controller->setInput(index++, floorSensor);
	}
	
	inline double sqr(double n)
	{
		return n*n;
	}
	
	double BatteryRecharge::simulateAmbiant(double xOffset, double yOffset, double angleOffset)
	{
		// compute sensor orientation and position
		double x, y, angle;
		x = pos.x + xOffset * cos(orientation);
		y = pos.y + yOffset * sin(orientation);
		angle = orientation + angleOffset;
		double angleDiff = M_PI + atan2(y, x) - angle; // angle difference between the sensor and the light (0;0)
		while(angleDiff > M_PI) angleDiff -= 2.0*M_PI;
		while(angleDiff < -M_PI) angleDiff += 2.0*M_PI;
		double r = sqrt(x*x + y*y);
		
		// simulate 
		double b = 25;	// ~ light intensity
		double s = 0.8;	// ~ angular field of view of the sensor
		
		return 1 / sqr(r / b + 1) * exp(-0.5 * sqr(angleDiff/s));
	}
	
	void BatteryRecharge::doIt()
	{
		#ifdef HAVE_VIEWER
		joystick = true;
		
		JoystickController js;
		robot->setPosition(An::Point(30.0, 30.0));
		robot->setOrientation(0.0);
		int i = 0;
		while (!js.isButtonDown(7))
		{
			// set robot speed and simulate
			double range, theta, speed;
			js.step();
			speed = js.getOutput(1);
			range = js.getOutput(3) / 5;
			theta = js.getOutput(2) / 5;
			leftJoystick = range * (speed + theta);
			rightJoystick = range * (speed - theta);
			
			// run experiment
			bool b;
			doSensoryMotorCycle(i++, b);
		}
		#else
		std::cerr << "You need to compile viewer to have joystick support." << std::endl;
		#endif
	}
}

#endif //--HAVE_ENKI
