#ifdef HAVE_ENKI
#include "KheperaNavigation.h"
#include "../utils/Logger.h"
#include "../controller/Joystick.h"
#include "../world/Enki.h"
#include "../main/VerboseLevel.h"
#include "../controller/ParameterGenotypeDecoder.h"
#include "../controller/Controller.h"

// Viewer stuff
#ifdef HAVE_VIEWER
#include "../enkiviewer/EnkiViewer.h"
#include "../enkiviewer/NeuralNetworkDrawer.h"
#endif

#include <an/Random.h>
#include <an/Types.h>

#include <enki/robots/khepera/Khepera.h>

#include <sstream>



namespace Teem
{
	#ifdef HAVE_VIEWER
	class KheperaNavigationDrawer
	{
	public:
		KheperaNavigationDrawer(KheperaNavigation* exp, GAG::GraphicContext *gc, int x, int y, int w, int h)
		{
			// draw IR values.
			int startX = x + 90;
			int startY = y + 37;
			int dw = 8;
			int dh = 40;
			int dx[] = {0, 2*dw, 4*dw, 6*dw, 8*dw, 10*dw, 6*dw, 4*dw};
			int dy[] = {dh, dh/2, 0, 0, dh/2, dh, dh*2, dh*2}; 
			for(unsigned i = 0; i < 8; i++)
			{
				gc->drawRect(startX + dx[i], startY + dy[i], dw, dh, 0, 0, 0);
				double d = dh * exp->ir[i];
				gc->drawFilledRect(startX + dx[i], (int)(startY + dy[i] + dh - d), dw, (int)d, 255, 255, 255);
			}
			
			// draw vision
			startX = x + 0;
			startY = y + 200;
			dw = 600 / exp->vision.size();
			if(dw > 8)
				dw = 8;
			dh = 30;
			for(unsigned i = 0; i < exp->vision.size(); i++)
			{
				unsigned c = static_cast<unsigned>(exp->vision[i] * 255.0);
				gc->drawFilledRect(startX + i*dw, startY, dw, dh, c, c, c);
				gc->drawRect(startX + i*dw, startY, dw+1, dh, 0, 0, 0);
			}
			
			// draw controller
			if(exp->handCodedMode)
			{
				NeuralNetworkDrawer nnd(gc, x, y + 250, 1000, 1000, 2.0);
				nnd.drawCell(0, 1, true);
				nnd.drawCell(0, 2, true);
				for(unsigned i = 0; i < 9; i++)
				{
					nnd.drawCell(i+1, 0, true);
					nnd.drawCell(i+1, 1, false, exp->leftWeights[i]);
					nnd.drawCell(i+1, 2, false, exp->rightWeights[i]);
				}
			}
			else
				NeuralNetworkDrawer(gc, x, y + 250, 1000, 1000, exp->controller, 2.0);
		}
	};
	
	class KheperaNavigationEnkiViewer : public VisionEnkiViewer
	{
	protected:
		KheperaNavigation* exp;
	public:
		KheperaNavigationEnkiViewer(const Enki::World *w, bool recVideo, KheperaNavigation* exp)
			: VisionEnkiViewer(w, recVideo, 600)
		{
			this->exp = exp;
		}
		virtual ~KheperaNavigationEnkiViewer() { }
		
	protected:
		
		virtual void displaySideBar(int x, int y, int w, int h)
		{
			VisionEnkiViewer::displaySideBar(x, y, w, h);
			KheperaNavigationDrawer(exp, gc, x, y, w, h);
		}
	};
	
	class KheperaNavigationViewer : public Viewer
	{
	protected:
		KheperaNavigation* exp;
	public:
		KheperaNavigationViewer(KheperaNavigation* exp) : Viewer(false)
		{
			this->exp = exp;
			init(600, 400);
		}
		
	protected:
		
		virtual void displayViewer(int w, int h)
		{
			gc->drawFilledRect(0, 0, w, h, 127, 127, 127);
			KheperaNavigationDrawer(exp, gc, 20, 20, w-40, h-40);
		}
	};
	
	#endif
	
	REGISTER_EXPERIMENT(KheperaNavigation);
	
	KheperaNavigation::KheperaNavigation() :
		inputIR("teem.experiment.inputIR", true),
		inputEncoders("teem.experiment.inputEncoders", true),
		inputVision("teem.experiment.inputVision", true),
		
		visionSize("teem.experiment.visionSize", 50),
		visionPreproc("teem.experiment.visionPreproc", 0),
		fieldOfView("teem.experiment.fieldOfView", 35),
		maxSpeed("teem.experiment.maxSpeed", 6.0),
		fitnessFunction("teem.experiment.fitnessFunction", "0.5 * (enc_l + enc_r)"),
		
		handCodedMode("teem.experiment.handCodedMode", false),
		leftWeights("teem.experiment.leftWeights", 0.0),
		rightWeights("teem.experiment.rightWeights", 0.0),
		
		fitnessFormula(fitnessFunction)
	{
		ir.resize(8);
		vision.resize(visionSize);
		joystick = false;
		
		// Create and setup world and robot.
		EnkiWorld* enkiWorld = new EnkiWorld(false);
		#ifdef HAVE_VIEWER
		enkiWorld->createViewer<KheperaNavigationEnkiViewer>(this);
		#endif
		w = enkiWorld;

		assert(w->getCapability<KheperaCreator>());
		robot = w->getCapability<KheperaCreator>()->createKhepera();
		
		// get enki instance of the robot
		EnkiKhepera* ek = dynamic_cast<EnkiKhepera*>(robot);
		if(ek != NULL)
		{
			enkiInstance = ek->getEnkiInstance();
			
			// init fov with the value in settings
			enkiInstance->camera.fieldOfView = fieldOfView / 180.0 * M_PI;
		}
		else
			enkiInstance = NULL;
			
		// Setup fitness formula.
		fitnessFormula.setVariable("enc_l", leftEncoder);
		fitnessFormula.setVariable("enc_r", rightEncoder);
		fitnessFormula.setVariable("mot_l", leftMotor);
		fitnessFormula.setVariable("mot_r", rightMotor);
		fitnessFormula.setVariable("x", position.x);
		fitnessFormula.setVariable("y", position.y);
		fitnessFormula.setVariable("orient", orientation);
		for(unsigned i = 0; i < 8; i++)
		{
			std::ostringstream oss; oss << "ir_" << i;
			fitnessFormula.setVariable(oss.str(), ir[i]);
		}
		for(unsigned i = 0; i < visionSize; i++)
		{
			std::ostringstream oss; oss << "vision_" << i;
			fitnessFormula.setVariable(oss.str(), vision[i]);
		}
		if(enkiInstance != NULL)
		{
			fitnessFormula.setVariable("fov", enkiInstance->camera.fieldOfView);
			fitnessFormula.setVariable("camDir", enkiInstance->camera.angleOffset);
		}
		
		fitnessFormula.parse();
	}
	
	KheperaNavigation::~KheperaNavigation()
	{
	}
		
	void KheperaNavigation::setupDecoders(GenotypeDecoders &decoders)
	{
		// count the number of inputs we are going to use
		unsigned numInput = 0;
		for(unsigned i = 0; i < 8; i++)
			if(inputIR[i])
				numInput++;
		if(inputEncoders[0]) numInput++;
		if(inputEncoders[1]) numInput++;
		if(inputVision) numInput += visionSize;
		
		// we might have a single decoder or 2, depending if the sensor morphology is evolved.
		assert(decoders.size() == 1 || decoders.size() == 2);
		decoders[0]->controllerIOSetup(numInput, 2);
	}
	
	void KheperaNavigation::createControllers(GenotypeDecoders &decoders, GenerationalPopulation &individuals, unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{
		assert(decoders.size() == 1 || decoders.size() == 2);
		assert(individuals.size() == 1 || individuals.size() == 2);
		
		controller = decoders[0]->decode(individuals[0].genome);
		
		// do we decode the parameters of the sensor?
		if(decoders.size() == 2)
		{
			assert(individuals.size() == 2);
			ParameterGenotypeDecoder* paramDecoder = dynamic_cast<ParameterGenotypeDecoder*>(decoders[1]);
			assert(paramDecoder);
			paramDecoder->decode(individuals[1].genome);
			
			// set the sensor parameters
			assert(enkiInstance != NULL);
			enkiInstance->camera.fieldOfView = paramDecoder->getParameter("fieldOfView") / 180.0 * M_PI;
			enkiInstance->camera.angleOffset = paramDecoder->getParameter("directionOfSight") / 180.0 * M_PI;
			
			COUT(3, "@param fieldOfView " << enkiInstance->camera.fieldOfView * 180.0 / M_PI << std::endl);
			COUT(3, "@param directionOfView " << enkiInstance->camera.angleOffset * 180.0 / M_PI << std::endl);
		}
	}
	
	void KheperaNavigation::initEpoch(unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{	
		// init robot position
		// we try to init the robot outside of an object
	
		EnkiWorld* ww = dynamic_cast<EnkiWorld*>(w);
		assert(ww);
		Enki::World* ew = ww->getEnkiInstance();
		robot->setPosition(An::Point(30, 30));
		robot->setOrientation(0);
		ew->step(0.00001);
		bool ok;
		double x, y;
		do
		{
			x = An::uniformRand() * 60.0;
			y = An::uniformRand() * 60.0;
			ok = true;
			for(size_t i = 0; i < ew->objects.size(); i++)
			{
				An::Vector unused;
				if(ew->objects[i]->boundingSurface != NULL)
					if(ew->isPointInside(An::Point(x, y), ew->objects[i]->pos, ew->objects[i]->absBoundingSurface, &unused))
					{
						ok = false;
						break;
					}
			}
		}
		while(!ok);
		robot->setPosition(An::Point(x, y));
		robot->setOrientation(An::uniformRand() * M_PI * 2.0);
		
		
		// setup logging
		if(mode == MODE_TEST_INDIVIDUAL)
		{
			std::ostringstream oss;
			oss << "epoch_ind" << ind << "_gen" << gen << ".log";
			logger = new OctaveLogger(oss.str().c_str(), true);
			logger->addVariable("leftEnc", leftEncoder);
			logger->addVariable("rightEnc", rightEncoder);
			logger->addVariable("leftMot", leftMotor);
			logger->addVariable("rightMot", rightMotor);
			logger->addVariable("x", position.x);
			logger->addVariable("y", position.y);
			logger->addVariable("orient", orientation);
			logger->addVariables("ir", &ir[0], &ir[ir.size()]);
			logger->addVariables("vision", &vision[0], &vision[vision.size()]);
			logger->addVariable("fit", instFitness);
			if(enkiInstance != NULL)
			{
				logger->addVariable("fov", enkiInstance->camera.fieldOfView);
				logger->addVariable("camDir", enkiInstance->camera.angleOffset);
			}
		}
		else
			logger = NULL;
	}
	
	void KheperaNavigation::termEpoch(unsigned gen, unsigned ind, unsigned epoch, EvaluationMode mode)
	{
		if(logger != NULL)
		{
			delete logger;
			logger = NULL;
		}
		
		delete controller;
	}
	
	double KheperaNavigation::doSensoryMotorCycle(unsigned cycle, bool &kill)
	{
		assert(robot != NULL);
		assert(controller != NULL || joystick);
		
		kill = false;
		
		// read sensors
		for(unsigned i = 0; i < 8; i++)
			ir[i] = robot->getIRSensor(i) / 1100.0;
		robot->getEncoders(leftEncoder, rightEncoder);
		position = robot->getPosition();
		leftEncoder /= maxSpeed;
		rightEncoder /= maxSpeed;
		readVisionSensor();
		
		// compute the fitness
		instFitness = computeFitness();
		
		// compute motor value
		if(handCodedMode)
		{
			runHandCodedNetwork(leftMotor, rightMotor);
		}
		else if(!joystick)
		{
			assignInputsToController();
			controller->step();
			leftMotor = controller->getOutput(0);
			rightMotor = controller->getOutput(1);
		}
		else
		{
			leftMotor = leftJoystick;
			rightMotor = rightJoystick;
		}
		
		// assign motor value and step
		robot->setSpeed(maxSpeed * leftMotor, maxSpeed * rightMotor);
		w->step(simulationStep);
		
		// log
		if(logger != NULL)
			logger->step();

		
		return instFitness;
	}
	
	double KheperaNavigation::computeFitness()
	{
		return fitnessFormula.eval();
	}
	
	void KheperaNavigation::assignInputsToController()
	{
		assert(controller != NULL);
		assert(robot != NULL);
		
		// ir sensors
		unsigned index = 0;
		for(unsigned i = 0; i < 8; i++)
			if(inputIR[i])
				controller->setInput(index++, ir[i]);
		
		// encoders
		if(inputEncoders[0]) controller->setInput(index++, leftEncoder);
		if(inputEncoders[1]) controller->setInput(index++, rightEncoder);
		
		// vision
		if(inputVision)
		{
			for(unsigned i = 0; i < vision.size(); i++)
				controller->setInput(index++, vision[i]);
		}
	}
	
	void KheperaNavigation::readVisionSensor()
	{
		switch(visionPreproc)
		{
			case RAW:
			{
				// subsample
				unsigned inc = 50 / visionSize;
				for(unsigned i = 0, j = 0; i < 50; i += inc, j++)
					vision[j] = robot->getCameraPixel(i).gray();
			}
			break;
			
			case SPATIAL_DIFF:
			{
				// subsample
				unsigned inc = 50 / visionSize;
				for(unsigned i = 0, j = 0; i < 50; i += inc, j++)
					vision[j] = robot->getCameraPixel(i).gray();
					
				// spatial difference
				for(unsigned i = 0; i < visionSize - 1; i++)
					vision[i] = fabs(vision[i] - vision[i+1]);
				vision[visionSize-1] = 0.0;
			}
			break;
			
			case DIFF_AVERAGE:
			{
				// spatial diff on the whole image
				std::valarray<double> img(50);
				for(unsigned i = 0; i < 49; i++)
					img[i] = fabs(robot->getCameraPixel(i).gray() - robot->getCameraPixel(i+1).gray());
				img[49] = 0.0;
				
				// averaging by chunk
				unsigned chunkSize = 50 / visionSize;
				for(unsigned i = 0; i < visionSize; i++)
				{
					double avg = 0.0;
					for(unsigned j = i*chunkSize; j < (i+1)*chunkSize; j++)
						avg += img[j];
					vision[i] = avg / chunkSize;
				}
			}
			break;
			
			default:
			assert(false && "Unsupported preprocessing");
			break;
		}
	}
	
	void KheperaNavigation::runHandCodedNetwork(double &left, double &right)
	{
		left = right = 0.0;
		
		// ir inputs
		for(unsigned i = 0; i < 8; i++)
		{
			left += ir[i] * leftWeights[i];
			right += ir[i] * rightWeights[i];
		}
		
		// bias
		left += -1.0 * leftWeights[8];
		right += -1.0 * rightWeights[8];
		
		// transfer function
		double b = 1.0;
		left = tanh(b * left);
		right = tanh(b * right);
	}
	
	void KheperaNavigation::doIt()
	{
		joystick = true;
		
		#ifdef HAVE_VIEWER
		JoystickController js;
		robot->setPosition(An::Point(30.0, 30.0));
		robot->setOrientation(0.0);
		int i = 0;
		while (!js.isButtonDown(7))
		{
			// set robot speed and simulate
			double range, theta, speed;
			js.step();
			speed = js.getOutput(1);
			range = js.getOutput(3) / 5;
			theta = js.getOutput(2) / 5;
			leftJoystick = range * (speed - theta);
			rightJoystick = range * (speed + theta);
			leftJoystick = 3;
			rightJoystick = 2;
			
			// run experiment
			bool b;
			doSensoryMotorCycle(i++, b);
		}
		#endif
	}
}
#endif // HAVE_ENKI
