/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef GENERATIONAL_EVOLUTION_H
#define GENERATIONAL_EVOLUTION_H

#include "Evolution.h"
#include <ishtar/settings.h>

#include <string>

/*!	\file GenerationalEvolution.h
	\brief Interfaces of algorithms for generations based evolution based
*/

namespace Ishtar
{
	class InputStream;
	class OutputStream;
}

namespace Teem
{
	class Genome;
	
	//! A individual for generational evolution
	//! \ingroup geneticalgos
	class GenerationalIndividual
	{
	public:
		Genome *genome; //!< Genome of this individual
		double fitness; //!< Fitness gained by this individual
		double trialCount; //!< Number of trial made by this individual
		
	public:
		//! Constructor
		GenerationalIndividual();
		//! Set fitness and trialCount to 0
		void reset(void);
		//! Divide fitness by trialCount
		void normalize(void);
		
		//! Load genome, fitness and trialCount from stream
		void load(Ishtar::InputStream *stream);
		//! Save gnome, fitness and trialCount to stream
		void save(Ishtar::OutputStream *stream);
	};
	
	//! Population of individual suited for generational evolution
	//! \ingroup geneticalgos
	typedef std::valarray<GenerationalIndividual> GenerationalPopulation;
	
	//! Interface to a class that implements a fitness function. Experiment
	//! classes for generational evolution typically derive from Evaluable.
	//! \ingroup geneticalgos
	class Evaluable : public Setupable
	{
	public:
		//!< Type of evaluation
		enum EvaluationMode
		{
			MODE_EVOLUTION = 0,	//!< running an evolution (do typically less logs for optimization)
			MODE_TEST_INDIVIDUAL	//!< testing an individual (do typically more logs for inspection)
		};
	
	public:
		//! Virtual destructor
		virtual ~Evaluable() { }
		
		//! Compute the fitness of the individual (or population).
		virtual void compute(GenotypeDecoders &decoders, GenerationalPopulation &individuals, unsigned gen, unsigned ind, EvaluationMode mode) = 0;
	};
	
	//! Comparison function to sort populations.
	//! \ingroup geneticalgos
	inline bool individualBacksort (const GenerationalIndividual & a, const GenerationalIndividual & b)
	{
		return a.fitness > b.fitness;
	}
	
	//! The basic class for generational GAs including all basic features. It is overloaded by specific algorithms.
	//! \ingroup geneticalgos
	class GenerationalEvolutionRun : public EvolutionRun
	{
	protected:
		// members
		//! A population, consisting of all individuals
		GenerationalPopulation population;
		//! The crossover probability per 
		Ishtar::Variable<double> crossoverProbability;
		//! The percentage of the population that is kept unchanged ("elite")
		Ishtar::Variable<double> elitismRatio; 
		//! The percentage of the population that will be reproduced. 
		Ishtar::Variable<double> reproductionRatio; 
		//! The total number of individuals present in the population
		Ishtar::Variable<unsigned> populationCount;
		//! The number of generations the experiment will run
		Ishtar::Variable<unsigned> generationCount;
		
		//! The current generation during runtime
		unsigned generation;
		//! Parameter used to toggle between evaluation and reproduction
		bool wasLoaded;
		
	protected:
		// support
		//! Delete unused genomes leftover from a previous generation
		void deleteGenomes(void);
		//! Get the population through Ishtar
		bool loadPopulationFromStream(Experiment* experiment, Ishtar::InputStream *stream);
		
		//! Compute the mean fitness of a group, which is a part of a population
		double computeMean(const std::valarray<GenerationalIndividual> &group);
		//! Compute the Standard Deviation of a group
		double computeSD(const std::valarray<GenerationalIndividual> &group);
		//! Calculate the Diversity of a group (we use the standard deviation of the Hamming distance)
		double computeDiversity(const std::valarray<GenerationalIndividual> &group);
		
		// steps for each generation
		//! The function used to test the population and determine a fitness for each individual
		virtual void evaluatePopulation(Evaluable *fitness) = 0;
		//! Randomly picks individuals from the part of the population defined by the reproductionRatio
		virtual void reproducePopulation(void) = 0;
		//! Documenting, data dumping and Writing to the log files
		virtual void logPopulation(void);
		//! Save the current population to the file
		virtual bool savePopulation(void);
		
		// wrapping
		//! Load a population at a specific generation
		virtual bool loadPopulation(Experiment* experiment, unsigned generation);
		//! Evaluate, Save, Log and Reproduce the population. Wrapping function
		virtual void run(Evaluable *fitness);
		//! Test the best individual and evaluate it (used to view and evaluate individuals for analysis). Wrapping function
		virtual void testBestIndividual(Evaluable *fitness) = 0;
		//! Test a certain individual and evaluate it (used to view and evaluate individuals for analysis). Wrapping function
		virtual void testIndividual(Evaluable *fitness, size_t number) = 0;
	
	public:
		//! Constuctor
		GenerationalEvolutionRun();
		//! Destructor
		virtual ~GenerationalEvolutionRun();
		
		// Main functions
		
		//! Create the necessary genotype decoders. Fill \c genotypeDecoders.
		//! @param experiment to provide a hook for experiment to configure decoders
		virtual void createGenotypeDecoders(Setupable *experiment) = 0;
		
		//! Create a population of random genomes, using the genotype decoder
		//! By default, use the first decoder to create the population.
		virtual void createRandomPopulation();
		
		//! Load a population from a filestream
		virtual bool loadPopulation(Experiment* experiment);

		//! Load a population at a specific generation
		virtual bool loadPopulation(Experiment* experiment, const char *generation) { return loadPopulation(experiment, atoi(generation)); }
		
		//! Evaluate, Save, Log and Reproduce the population
		virtual void run(Setupable *experiment);
		
		//! Test the best Individual and test it (used to view and evaluate indivuals for analysis)
		virtual void testBestIndividual(Setupable *experiment);
		//! Test a certain individual and evaluate it (used to view and evaluate individuals for analysis)
		virtual void testIndividual(Setupable *experiment, const char *);
	};
	
	//! Base class for evolution runs that need only one genotype decoder. That is typically the case for runs where only one type of robot exist.
	//! \ingroup geneticalgos
	class SingleDecoderEvolutionRun : public GenerationalEvolutionRun
	{
	public:
		//! Constructor
		SingleDecoderEvolutionRun() : genotypeDecoderName("teem.genotypeDecoderName", "FeedForward") { }
		//! Virtual destructor, doing nothing
		virtual ~SingleDecoderEvolutionRun() { }
		
	protected:
		//! Name of the genotype decoder to be used.
		Ishtar::Variable<std::string> genotypeDecoderName;
		
		//! Create the necessary genotype decoders. Fill \c genotypeDecoders.
		virtual void createGenotypeDecoders(Setupable *experiment);
	};
	
	//! Simplest possible evolution run. To be used for single individual experiment.
	//! \ingroup geneticalgos
	class SimpleEvolutionRun: public SingleDecoderEvolutionRun
	{
	protected:
		//! Evaluate the whole population and determine individual fitness
		virtual void evaluatePopulation(Evaluable *fitness);
		//! Reproduce the population according to the parameters reproductionRatio, elitismRatio and crossoverProbability and mutate according to mutate() in Genome
		virtual void reproducePopulation(void);
		
		virtual void testBestIndividual(Evaluable *fitness);
		virtual void testIndividual(Evaluable *fitness, size_t number);
		
	public:
		//! A simple evolution 
		SimpleEvolutionRun();
	};
	
	//! Base class for evlution runs that need many different genome and decoders for each individual.
	//! \ingroup geneticalgos
	class MultiDecoderEvolutionRun : public GenerationalEvolutionRun
	{
	public:
		//! Constructor
		MultiDecoderEvolutionRun();
		//! Virtual destructor, doing nothing
		virtual ~MultiDecoderEvolutionRun() { }
		
	protected:
		//! Number of genotypeDecoder to be used.
		Ishtar::Variable<unsigned> numGenotypeDecoder;
		
		//! Name of the genotyeDecoder (registry ID).
		std::vector<Ishtar::Variable<std::string> > genotypeDecoderNames;
		
		//! Create the necessary genotype decoders. Fill \c genotypeDecoders.
		virtual void createGenotypeDecoders(Setupable *experiment);
	};
	
	//! Simple evoluation run. Conceptually it is single individual, but multiple
	//! genomes and decoders are passed to the experiment, so it is effectively
	//! multi genome individuals. The experiment needs only to set the fitness of
	//! the FIRST genome.
	//! All the genome are stored in a single population and are ordered by individual,
	//! that is the N genome for the first individual come first, then the N for the
	//! second individual, etc.
	//! \ingroup geneticalgos
	class SimpleMultiGenomeEvolutionRun :  public MultiDecoderEvolutionRun
	{
	protected:
		//! Evaluate the whole population and determine individual fitness
		virtual void evaluatePopulation(Evaluable *fitness);
		//! Reproduce the population according to the parameters reproductionRatio, elitismRatio and crossoverProbability and mutate according to mutate() in Genome
		virtual void reproducePopulation(void);
		//! Create a new random population.
		virtual void createRandomPopulation();
		
		virtual void testBestIndividual(Evaluable *fitness);
		virtual void testIndividual(Evaluable *fitness, size_t number);
		
	public:
		//! A simple evolution 
		SimpleMultiGenomeEvolutionRun();
	};
}

#endif
