#ifndef __GRAPH_H
#define __GRAPH_H

#include <string>
#include <vector>
#include <set>
#include <map>
#include <ishtar/settings.h>

/*!	\file Graph.h
	\brief Interface of classes to create graphs
*/

namespace Teem
{
	/**
	Utility class to provide a nice syntax for node/link creation.
	Ex: graph.createNode("nodeType", TagSet("output", "special"));
	Is a std::set of std::string with useful accessors
	\ingroup utils
	*/
	class TagSet : public std::set<std::string>
	{
	public:
		//! Destroy the set of strings
		virtual ~TagSet() {}
		//! Constructor, create an empty set of strings
		TagSet() {}
		//! Constructor, create a set and insert the initial string tag
		TagSet(const std::string &tag) { insert(tag); }
		//! Add string tag to the string set
		TagSet& operator+(const std::string &tag) { insert(tag); return *this; }
		//! Add string tag to the string set
		TagSet& operator+(const char *tag) { insert(tag); return *this; }
	};
	
	/**
	Utility class that will autmatically fill itself with settings
	from Ishtar. The root setting must contain only single value,
	numerical Ishtar parameters.
	\ingroup utils
	*/
	class IshtarSettingsMap : public std::map<std::string, double>
	{
	public:
		//! Constructor
		IshtarSettingsMap(const std::string &root);
		//! Destructor
		virtual ~IshtarSettingsMap() { }
	
	protected:
		/*//! Comparaison functor
		class Comparator
		{
		protected:
			const std::string &key;
			
		public:
			Comparator(const std::string &theKey) : key(theKey) { }
			int operator()(const std::pair<std::string, Ishtar::ParsedValues> &first, const std::pair<std::string, Ishtar::ParsedValues> &second)
			{
				if (first.first.compare("") == 0)
					return second.first.find(key) == 0;
				else
					return first.first.find(key) == 0;
			}
		};*/
	};
	
	/**
	Class to build and describe a generalized graph connecting nodes with links.
	Each node and link has a list of (key; value) property. Nodes and links also
	have a \c type parameter that describe the type of object it is. This typically
	corespond to some registery ID to create a structure based on a graph.
	
	\see GraphNeuralNetwork
	
	\ingroup utils
	*/
	class Graph
	{
		friend class GraphNeuralNetwork;
		
	public:
		//! A Node is an id in nodes
		typedef size_t Node;
		//! A Link is an id in links
		typedef size_t Link;
		
		//! Interface class for object that want to hook into a Graph before
		//! conversion to some controller.
		//! @see GraphGenotypeDecoder
		class Filter
		{
		public:
			//! Virtual destructor do nothing.
			virtual ~Filter() { }
			
			//! Apply this filter the a graph
			virtual void filter(Graph *graph) = 0;
		};
		
	public:
		//! Constructor
		Graph() { }
		//! Destructor. All datas are ereased because they are all values not pointers
		virtual ~Graph() { }
		
		//! Add a param that will be added to newly create node.
		void addDefaultNodeParam(const std::string &id, double value) { defaultNodeParams[id] = value; }
		
		//! Add params that will be added to newly create node.
		void addDefaultNodeParam(const std::map<std::string, double> &params) { defaultNodeParams.insert(params.begin(), params.end()); }

		//! Add a param that will be added to newly created links.
		void addDefaultLinkParam(const std::string &id, double value) { defaultLinkParams[id] = value; }
		
		//! Add params that will be added to newly created links.
		void addDefaultLinkParam(const std::map<std::string, double> &params) { defaultLinkParams.insert(params.begin(), params.end()); }
		
		/**
		Create a new node.
		@param type type of the node (as registered)
		@param tags set of tag for the node
		*/
		Node createNode(const std::string &type, const std::set<std::string> &tags = TagSet());
		
		/**
		Create a new link.
		@param from start node for the link
		@param to destination node for the link
		@param type type of the link (as registered)
		@param tags set of tag for the link
		*/
		Link createLink(Node from, Node to, const std::string &type, const std::set<std::string> &tags = TagSet());
		
		//! Add a new parameter to the node.
		void addNodeParam(Node node, const std::string &id, double value) {nodes[node].params[id] = value;}
		
		//! Add new parameters to the node.
		void addNodeParam(Node node, const std::map<std::string, double> &params) {nodes[node].params.insert(params.begin(), params.end());}
		
		//! Add a new parameter to the link.
		void addLinkParam(Link link, const std::string &id, double value) {links[link].params[id] = value;}
		
		//! Add new parameters to the link.
		void addLinkParam(Link link, const std::map<std::string, double> &params) {links[link].params.insert(params.begin(), params.end());}
		
		//! Get a node using an index list.
		Node getNode(const std::string &tag, size_t index) { return nodeLists[tag][index]; }
		
		//! Get the number of node in a tagged list.
		size_t getNodeCount(const std::string &tag) { return nodeLists[tag].size(); }
		
		//! Get a link using an index list.
		Link getLink(const std::string &tag, size_t index) { return linkLists[tag][index]; }
		
		//! Get the number of link in a tagged list.
		size_t getLinkCount(const std::string &tag) { return linkLists[tag].size(); }
		
		//! Export the graph to dot format.
		void exportToDot(const std::string &filename) const;
		
		//! Export to graph to a Matlab compatible file that describe its structure.
		//! Note that all indexes are converted to be one-based.
		void exportToMatlab(const std::string &filename) const;
		
	protected:
		
		//! Data structure that describe a node.
		struct NodeData
		{
			std::string type; //!< type of node
			std::set<std::string> tags; //!< tags of node
			std::map<std::string, double> params; //!< names and values of node datas
		};
		
		//! Data structure that describe a link.
		struct LinkData
		{
			std::string type; //!< type of link
			std::set<std::string> tags; //!< tags of link
			Node from; //!< source node
			Node to; //!< destination node
			std::map<std::string, double> params; //!< names and values of link datas
		};
		
		std::vector<NodeData> nodes; //!< nodes content
		std::vector<LinkData> links; //!< links content
		
		std::map<std::string, double> defaultNodeParams; //!< Default node parameters.
		std::map<std::string, double> defaultLinkParams; //!< Default link parameters. 
		
		std::map<std::string, std::vector<Node> > nodeLists; //!< Index lists for nodes.
		std::map<std::string, std::vector<Link> > linkLists; //!< Index lists for links.

	};
}

#endif
