#ifndef __SPIKEDELAYLINE_H
#define __SPIKEDELAYLINE_H

#include <queue>

/*!	\file SpikeDelayLine.h
	\brief Interface of a spike delay line
*/

namespace Teem
{
	//! An helper class to store spikes for a certain duration
	//! \ingroup utils
	class SpikeDelayLine
	{
	public:
		//! Delay of the line
		SpikeDelayLine(double theDelay = 2.0);
		//! Destructor
		virtual ~SpikeDelayLine() {}
		
		//! Step the delay line.
		//! @param dt size of time step.
		//! @param pushSpike set to true if a spike should be pushed in the line durint this time-step
		//! @return the number of spikes that exit the line during this time-step
		unsigned step(double dt, bool pushSpike);
		
		//! Set a new value for the delay. In case the delay is reduced, it is possible to 
		//! dump spikes that are beyond the new delay immediately or to wait for the next 
		//! time-step. In the last case, the function always returns 0.
		//! @param theDelay new value for the delay
		//! @param dumpSpikes should the old spikes be dumped
		//! @return number of spike that are dumped (in case the delay is reduced)
		unsigned setDelay(double theDelay, bool dumpSpikes = true);
		//! Get the current delay.
		double getDelay() { return delay; }
		
		//! Returns there is one (or more) spike in the delay line.
		bool spikeInLine() const { return spikePresent; }
		//! Returns the number of spike in the delay line.
		unsigned spikeCount() const { return (intervals.size() == 0) ? (spikePresent ? 1 : 0) : intervals.size() + 1; }
	
	protected:
	
		//! Length of the delay line.
		double delay;
		//! Current total time the in the intervals. 
		double totalInterval;
		//! Time since last spike pushed in the line.
		double timeSinceLastSpike;
		//! True if spike are in line.
		bool spikePresent;
		//! Inter-spike intervals.
		std::queue<double> intervals;
		
		//! Pop all the spikes that are beyond the current delay.
		//! @return the number of spikes that were removed.
		unsigned dumpOldSpikes();
	};
}

#endif
