/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __TEEM_ENTITY_H
#define __TEEM_ENTITY_H

#include "EntityInterfaces.h"
#include <an/Types.h>

/*!	\file Entity.h
	\brief Interfaces of entities that can exist in worlds
*/

namespace Teem
{
	//! This is the interface to a Khepera.
	//! \ingroup world
	class Khepera :
		public Entity,
		virtual public Positionable,
		public DifferentialWheeled,
		public IRSensorEquipped,
		public Camera1DEquipped
	{
	protected:
		friend class KheperaCreator;
		//! Protected constructor, created by KheperaCreator-capable worlds.
		Khepera() { }

	public:
		virtual ~Khepera() { }
		
		// TODO: should be in differentialWheeled, but not implemented in other robots yet
		//! Return the encoder values in left and right.
		virtual void getEncoders(double &left, double &right) = 0;
	};
	
	//! This is the interface to a Blimp robot.
	//! \ingroup world
	class Blimp :
		public Entity,
		public Camera1DEquipped
	{
	protected:
		friend class BlimpCreator;
		//! Protected constructor, created by BlimpCreator-capable worlds.
		Blimp() { }
	
	public:
		virtual ~Blimp() { }
		
		//! Set the speed of the Blimp to forward and orientation to dir.
		virtual void setSpeed(double forward, double dir) = 0;
		
		//! Return the value of the gyroscope (measured rotation speed).
		virtual double getGyro() = 0;
		//! Return the value of the anemometer (measured forward speed).
		virtual double getAnemo() = 0;
		//! Return the value of the bumper index (collision detection).
		virtual bool getBumper(unsigned index) = 0;
		//! Return the position of the Blimp.
		virtual const An::Point &getPosition() = 0;
	};
	
	//! This is the interface to an Alice.
	//! \ingroup world
	class Alice :
		public Entity,
		virtual public Positionable,
		public DifferentialWheeled,
		public IRSensorEquipped,
		public Camera1DEquipped
	{
	protected:
		friend class AliceCreator;
		//! Protected constructor, created by AliceCreator capable worlds.
		Alice() { }

	public:
		virtual ~Alice() { }
		
		//! Broadcast value val using IR transmission.
		virtual void setTransmitValue(unsigned val) = 0;
		//! Enable/disable IR transmission.
		virtual void setTransmit(bool enabled) = 0;
		//! Return true if a value has been received by IR transmission.
		virtual bool wasCommunication(void) = 0;
		//! Return the sensor on which the last value was received.
		virtual unsigned getSensorId(void) = 0;
		//! Return the last value received.
		virtual unsigned getSensorValue(void) = 0;
		
		//! Return the speed of Alice wheels in left and right.
		virtual void getSpeed(double &left, double &right) = 0;
	};
	
	//! This is the interface to a small circular disk present in Alice worlds.
	//! \ingroup world
	class AliceSmallDisk :
		public Entity,
		virtual public Positionable
	{
	protected:
		friend class AliceSmallDiskCreator;
		//! Protected constructor, created by AliceSmallDiskCreator-capable worlds.
		AliceSmallDisk() { }
	};
	
	//! This is the interface to a large circular disk present in Alice worlds.
	//! \ingroup world
	class AliceLargeDisk :
		public Entity,
		virtual public Positionable
	{
	protected:
		friend class AliceLargeDiskCreator;
		//! Protected constructor, created by AliceLargeDiskCreator-capable worlds.
		AliceLargeDisk() { }
	};

	//! This is the interface to an Sbot.
	//! \ingroup world
	class Sbot :
		public Entity,
		virtual public Positionable,
		public DifferentialWheeled
	{
	protected:
		friend class SbotCreator;
		//! Protected constructor, created by SbotCreator-capable worlds.
		Sbot() { }
	
	public:
		virtual ~Sbot() { }
		//! Sbot camera orientation is geometric (i.e. CCW).
		virtual void getCameraAverage(An::Color meanPixels[8]) = 0;
		//! Sbot camera orientation is geometric (i.e. CCW).
		virtual void getCamera(An::Color meanPixels[128]) = 0;
		//! Set the color of the Sbot ring in RGB using range [0..1].
		virtual void setRingColor(const An::Color &color) = 0;
		
		//! Clear the energy of this Sbot, equal to setEnergy(0).
		virtual void clearEnergy(void) = 0;
		//! Set the energy of this Sbot to energy.
		virtual void setEnergy(double energy) = 0;
		//! Return the actual energy of this Sbot.
		virtual double getEnergy(void) = 0;
		//! Return the actual energy derivate (one step increment) of this Sbot.
		virtual double getDEnergy(void) = 0;
		
		//! Transmit sound using frequencies mask.
		virtual void setFrequencies(unsigned frequencies) = 0;
		//! Return a mask of active frequencies in the world.
		virtual unsigned getFrequencies(void) = 0;
	};

	//! This is the interface to an Sbot active object.
	//! \ingroup world
	class SbotActiveObject :
		public Entity,
		virtual public Positionable
	{
	protected:
		friend class SbotCreator;
		//! Protected constructor, created by SbotCreator-capable worlds.
		SbotActiveObject() { }

	public:
		virtual ~SbotActiveObject() { }
		
		//! Set the duration of the two states of this object: active and inactive. Set energy consumption on and off as well.
		virtual void setEnergyProperties(double activeDuration, double inactiveDuration, bool consumeEnergy) = 0;
		//! Set the actual time in energy cycle.
		virtual void setEnergyTime(double time) = 0;
		//! Set the energy derivate to nearby Sbots while in active/inactive states.
		virtual void setEnergyFlow(double activeFlow, double inactiveFlow) = 0;
		//! Set the energy of this object.
		virtual void setEnergy(double energy) = 0;
		//! Return the energy of this object.
		virtual double getEnergy(void) = 0;
		
		//! Set the color of this object for both active and inactive states. Values are RGB in range [0..1].
		virtual void setColor(const An::Color &activeColor, const An::Color &inactiveColor) = 0;
		
		//! Return the range up to which this object interacts with Sbots.
		virtual double getInteractionRange(void) = 0;
	};
	
	//! This is an interface to a square piece of static matter.
	//! \ingroup world
	class RectangularBlock :
		public Entity,
		virtual public Positionable
	{
	protected:
		friend class RectangularBlockCreator;
		//! Protected constructor, created by RectangularBlockCreator-capable worlds.
		RectangularBlock() { }
	
	public:
		virtual ~RectangularBlock() { }
		//! Set the width, height, position and orientation of this rectangular block.
		virtual void set(double width, double height, double x, double y, double orientation) = 0;
	};
}

#endif
