/*
Teem - an open source evolutionary software framework
http://lis.epfl.ch/resources/Teem
Stephane Magnenat <stephane.magnenat@epfl.ch>,
Antoine Beyeler <antoine.beyeler@epfl.ch>,
and other contributors. See AUTHORS for details
Laboratory of Intelligent Systems, EPFL, Lausanne

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __TEEM_WORLD_H
#define __TEEM_WORLD_H

#include <an/Types.h>
#include <an/Geometry.h>

/*!	\file World.h
	\brief Generic capabilities of worlds
*/

/*!	\defgroup world World interfaces
	The interfaces and glue code to worlds, simulated and reals
*/

namespace Teem
{
	//! A capability of a world
	//! \ingroup world
	class WorldCapability
	{
	public:
		//! Virtual destructor, do nothing.
		virtual ~WorldCapability() { }
	};

	class Sbot;

	//! Can create an Sbot.
	//! \ingroup world
	class SbotCreator : public WorldCapability
	{
	public:
		virtual ~SbotCreator() { }
		//! Create an Sbot.
		virtual Sbot *createSbot() = 0;
	};

	class SbotActiveObject;

	//! Can create an Sbot.
	//! \ingroup world
	class SbotActiveObjectCreator : public WorldCapability
	{
	public:
		virtual ~SbotActiveObjectCreator() { }
		//! Create an S-bot active object.
		virtual SbotActiveObject *createSbotActiveObject() = 0;
	};

	class Alice;

	//! Can create an Alice.
	//! \ingroup world
	class AliceCreator : public WorldCapability
	{
	public:
		virtual ~AliceCreator() { }
		//! Create an Alice.
		virtual Alice *createAlice() = 0;
	};
	
	class AliceSmallDisk;

	//! Can create an AliceDisk.
	//! \ingroup world
	class AliceSmallDiskCreator : public WorldCapability
	{
	public:
		virtual ~AliceSmallDiskCreator() { }
		//! Create a small disk suitable for Alice related experiments.
		virtual AliceSmallDisk *createAliceSmallDisk() = 0;
	};
	
	class AliceLargeDisk;

	//! Can create an AliceDisk.
	//! \ingroup world
	class AliceLargeDiskCreator : public WorldCapability
	{
	public:
		virtual ~AliceLargeDiskCreator() { }
		//! Create a large disk suitable for Alice related experiments.
		virtual AliceLargeDisk *createAliceLargeDisk() = 0;
	};
	
	class AliceBlock;
	
	//! Can create an AliceBlock.
	//! \ingroup world
	class AliceBlockCreator: public WorldCapability
	{
	public:
		virtual ~AliceBlockCreator() {}
		//! Create a block suitable for Alice related experiments.
		virtual AliceBlock *createAliceBlock(bool high) = 0;
	};
	
	class Khepera;

	//! Can create a Khepera.
	//! \ingroup world
	class KheperaCreator : public WorldCapability
	{
	public:
		virtual ~KheperaCreator() { }
		//! Create a Khepera.
		virtual Khepera *createKhepera() = 0;
	};
	
	class Blimp;
	//! Can create a Blimp. Not supported by the Enki backend but will be supported by other backends to be implemented.
	//! \ingroup world
	class BlimpCreator : public WorldCapability
	{
	public:
		virtual ~BlimpCreator() { }
		//! Create a blimp.
		virtual Blimp* createBlimp() = 0;
	};
	
	class RectangularBlock;
	
	//! Can create blocks.
	//! \ingroup world
	class RectangularBlockCreator : public WorldCapability
	{
	public:
		virtual ~RectangularBlockCreator() { }
		//! Create a rectangular block of matter.
		virtual RectangularBlock *createRectangularBlock() = 0;
	};
	
	
	//! This is the interface to the world, mother of existence, purely virtual of course.
	//! \ingroup world
	class World
	{
	public:
		//! Virtual destructor, do nothing.
		virtual ~World() { }
		//! If this world has capability C, return it; otherwise, return NULL.
		template<class C>
		C *getCapability(void) { return dynamic_cast<C *>(this); }
		//! Run the world for dt milliseconds.
		virtual void step(double dt) = 0;
	};
	
	//! This is the interface to any arena world (bounded by walls).
	//! \ingroup world
	class ArenaWorld : public World
	{
	public:
		virtual ~ArenaWorld() { }
	};
	
	//! This is the interface to a square world.
	//! \ingroup world
	class SquareArenaWorld : public ArenaWorld
	{
	public:
		virtual ~SquareArenaWorld() { }
		//! Return the size of this square world.
		virtual An::Point getDimension() = 0;
		//! Set the wall textures of this world.
		virtual void setWallTextures(const std::valarray<An::Texture> &textures) = 0;
		//! Enable/disable the use of walls. A square arena has walls by default.
		virtual void useWalls(bool use) = 0;
	};
}

#endif
